import logging
import os
import threading
from typing import Union, Callable

from sqlalchemy.pool import QueuePool
from sqlalchemy.util import queue


def __evaluate__(_callable_or_value):
    _cv = _callable_or_value
    return _cv() if callable(_cv) else _cv


class HashableDict(dict):

    def __hash__(self):
        return hash(frozenset(self.items()))


class IntervalTimer(threading.Thread):

    def __init__(self, interval, func, args=None, kwargs=None):
        """
        :type interval: float or Callable
        :type func: Callable
        :type args: list or tuple
        :type kwargs: dict
        """
        threading.Thread.__init__(self)
        self.interval = interval  # type: Union[float, Callable[[], float]]
        self.function = func
        self.args = args or []
        self.kwargs = kwargs or {}
        self.finished = threading.Event()

    def cancel(self):
        self.finished.set()

    def run(self):
        while True:
            interval = __evaluate__(self.interval)
            if self.finished.wait(interval):
                break
            self.function(*self.args, **self.kwargs)
        self.finished.set()


class DjangoQueuePool(QueuePool):
    def __init__(
            self,
            creator,
            core_pool_size=None,
            unload_timeout=None,
            retire_interval=None,
            retire_quantity=None,
            show_status_interval=None,
            **kwargs
    ):
        """
        :type creator: Callable
        :type core_pool_size: int or Callable
        :type unload_timeout: int or float or Callable
        :type retire_interval: int or float or Callable
        :type retire_quantity: int or float or Callable
        :type kwargs: dict
        """
        super(DjangoQueuePool, self).__init__(creator, **kwargs)

        if self.logging_name and kwargs.get("echo"):
            self.logger = logging.getLogger(self.logging_name)

        self._core_pool_size = core_pool_size  # type: Union[int, Callable[[], int]]

        self._unload_timeout = unload_timeout  # type: Union[float, Callable[[], float]]

        self._retire_timer = None
        self._retire_interval = retire_interval  # type: Union[float, Callable[[], float]]
        self._retire_quantity = retire_quantity or 1  # type: Union[int, Callable[[], int]]

        if self._retire_interval is not None:
            self._retire_timer = IntervalTimer(self._retire_interval, self._do_retire)
            self._retire_timer.daemon = True
            self._retire_timer.start()

        self._show_status_timer = None
        self._show_status_interval = show_status_interval

        if self._show_status_interval is not None:
            self._show_status_timer = IntervalTimer(self._show_status_interval, self.status)
            self._show_status_timer.daemon = True
            self._show_status_timer.start()

    def _is_overload(self):
        return self.checkedout() >= __evaluate__(self._core_pool_size)

    def _do_get(self):

        if (
                self._is_overload() and
                self._unload_timeout is not None
        ):
            try:  # wait before overflow
                return self._pool.get(
                    True, __evaluate__(self._unload_timeout))
            except queue.Empty:
                pass

        return super(DjangoQueuePool, self)._do_get()

    # noinspection PyBroadException
    def do_retire_conn(self):
        try:
            conn = self._do_get()
            if conn is not None:
                try:
                    conn.close()
                finally:
                    self._dec_overflow()
        except Exception:
            pass

    # noinspection PyBroadException
    def _do_retire(self):
        try:
            for _ in range(__evaluate__(self._retire_quantity)):
                self.do_retire_conn()
        except Exception:
            pass

    def recreate(self):

        return self.__class__(

            self._creator,

            core_pool_size=self._core_pool_size,
            unload_timeout=self._unload_timeout,
            retire_interval=self._retire_interval,
            retire_quantity=self._retire_quantity,

            pool_size=self._pool.maxsize,
            max_overflow=self._max_overflow,
            timeout=self._timeout,
            recycle=self._recycle,
            dialect=self._dialect,
            pre_ping=self._pre_ping,
            reset_on_return=self._reset_on_return,
            echo=self.echo,
            logging_name=self._orig_logging_name,
            _dispatch=getattr(self, 'dispatch', None),
        )

    def core_size(self):
        return self._core_pool_size

    def status(self):
        pid = os.getpid()
        status = (
            f"Mysql Connection Pool of {pid}; "
            f"Pool size: {self.size()}; Connections in pool: {self.checkedin()}; "
            f"Current Overflow: {self.overflow()}; Current Checked out: {self.checkedout()}"
        )
        self.logger.info(status)
        return status



"""
Django ORM 补丁, 执行完查询后立即释放连接
"""

from django.db.models.sql import compiler
from django.db.models.sql.constants import MULTI, GET_ITERATOR_CHUNK_SIZE


def install_django_orm_patch():
    execute_sql = compiler.SQLCompiler.execute_sql

    def patched_execute_sql(self, result_type=MULTI, chunked_fetch=False, chunk_size=GET_ITERATOR_CHUNK_SIZE):
        try:
            result = execute_sql(self, result_type, chunked_fetch, chunk_size)
        finally:
            if not self.connection.in_atomic_block:
                self.connection.close()  # return connection to pool by db_pool_patch
        return result

    compiler.SQLCompiler.execute_sql = patched_execute_sql

    insert_execute_sql = compiler.SQLInsertCompiler.execute_sql

    def patched_insert_execute_sql(self, return_id=False):
        try:
            result = insert_execute_sql(self, return_id)
        finally:
            if not self.connection.in_atomic_block:
                self.connection.close()  # return connection to pool by db_pool_patch
        return result

    compiler.SQLInsertCompiler.execute_sql = patched_insert_execute_sql



import collections
import os
import threading

from django.db.backends.mysql import base as mysql
from django.db.utils import DatabaseError

try:
    from django.db.backends.mysql.base import Database
except ImportError as e:
    raise ImportError('Error loading DB-API 2.0 module: %s' % e)

from .pool import HashableDict

import copy

from django.utils import module_loading
from sqlalchemy.pool import Pool


class PoolParams(object):
    def __init__(self, pool_class: str = None, **kwargs):
        self.pool_class = module_loading.import_string(pool_class)
        if not issubclass(self.pool_class, Pool):
            raise Exception('`pool_class` must be a subclass of sqlalchemy.pool.Pool or a import string of it')

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_pool(self, creator):
        params = copy.copy(self.__dict__)
        params.pop('pool_class', None)
        return self.pool_class(creator, **params)


class DatabaseWrapper(mysql.DatabaseWrapper):
    db_proxy_lock = threading.Lock()
    db_proxy_by_pid = dict()
    _pool_params = None

    def creator(self):
        conn_params = self.get_connection_params()
        return Database.connect(**conn_params)

    def create_db_proxy(self):
        pool = self.pool_params.get_pool(self.creator)
        return pool

    @property
    def db_proxy(self):
        cls = self.__class__
        pid = os.getpid()
        db_name = self.settings_dict.get('NAME')
        db_proxy_key = f"{self.alias}_{db_name}"
        db_proxy_by_pid = cls.db_proxy_by_pid
        db_proxy = db_proxy_by_pid.get(pid, {})
        db_proxy = db_proxy.get(db_proxy_key)

        if db_proxy is not None:
            return db_proxy
        with cls.db_proxy_lock:
            db_proxy = db_proxy_by_pid.get(pid)

            if db_proxy is None:
                db_proxy_by_pid[pid] = {}
                db_proxy = db_proxy_by_pid[pid]

            db_proxy = db_proxy.get(db_proxy_key)
            if db_proxy is not None:
                return db_proxy

            db_proxy = self.create_db_proxy()

            db_proxy_by_pid[pid].update({
                db_proxy_key: db_proxy
            })

        if db_proxy is None:
            raise Exception('unable to initialize the database proxy')
        return db_proxy

    @property
    def pool_params(self):
        """
        :rtype: DjangoPoolParams
        """
        if self._pool_params is None:
            self._pool_params = PoolParams(**self.settings_dict['POOL'])
        return self._pool_params

    def get_connection_params(self):
        raw_params = super(DatabaseWrapper, self).get_connection_params()
        new_params = dict()
        for raw_key, raw_value in raw_params.items():
            new_value = raw_value
            if not isinstance(raw_value, collections.Hashable):
                if isinstance(raw_value, dict):
                    new_value = HashableDict(raw_value)
                elif isinstance(raw_value, list):
                    new_value = tuple(raw_value)
                elif isinstance(raw_value, set):
                    new_value = frozenset(raw_value)
            if not isinstance(new_value, collections.Hashable):
                raise Exception('unhashable connection parameter %s' % raw_key)
            new_params[raw_key] = new_value
        return new_params

    def get_new_connection(self, conn_params):
        return self.db_proxy.connect()

    def connect(self):

        super(DatabaseWrapper, self).connect()

        if not (
                self.connection is not None and
                self.pool_params.pre_ping
        ):
            return

        is_usable = False
        ex_message = None
        ex_default = 'unable to connect to the database'
        try:
            is_usable = self.is_usable()  # ping
        except Exception as ex:
            ex_message = str(ex)
        finally:
            if not is_usable:
                try:
                    self.errors_occurred = True
                    self.close()
                finally:
                    raise DatabaseError(ex_message or ex_default)

    def _close(self):

        conn = self.connection  # type: sqlalchemy.pool._ConnectionFairy

        if conn is None:
            return

        with self.wrap_database_errors:
            try:
                if self.in_atomic_block:
                    if self.pool_params.reset_on_return:
                        conn.rollback()
            finally:
                if self.errors_occurred:
                    conn.invalidate()
                else:
                    conn.close()



from setuptools import setup, find_packages

name = 'django-connection-pool'
version = '1.2.0'

setup(
    name=name,
    version=version,
    packages=find_packages(),
    install_requires=[
        "Django>=3.0.0,<4.0.0",
        "sqlalchemy>=1.4.20",
    ],
    python_requires='>=3.7',
    include_package_data=True,
    license='GNU General Public License v3.0',
    author='Lee',
    author_email='canyun@live.com',
    description='django connection pool',
)
